package com.tzld.piaoquan.recommend.server.service.rank;

import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.RandomUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;

import java.util.*;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
public class RankService {
    @Autowired
    private RedisTemplate<String, String> redisTemplate;

    public RankResult rank(RankParam param) {

        if (param.getRecallResult() == null
                || (CollectionUtils.isEmpty(param.getRecallResult().getRovPoolRecall())
                && CollectionUtils.isEmpty(param.getRecallResult().getFlowPoolRecall()))) {
            return null;
        }

        // rank rov recall
        List<String> videoIdKeys = param.getRecallResult().getRovPoolRecall().stream()
                .map(t -> param.getRankKeyPrefix() + t.getVideoId())
                .collect(Collectors.toList());
        List<RecallResult.RecallData> rov_recall_rank = param.getRecallResult().getRovPoolRecall().stream()
                .collect(Collectors.toList());
        List<String> video_scores = redisTemplate.opsForValue().multiGet(videoIdKeys);
        if (CollectionUtils.isNotEmpty(video_scores)
                && video_scores.size() == rov_recall_rank.size()) {
            for (int i = 0; i < video_scores.size(); i++) {
                rov_recall_rank.get(i).setRovScore(NumberUtils.toDouble(video_scores.get(i), 0.0));
            }
            Collections.sort(rov_recall_rank, Comparator.comparingDouble(o -> -o.getRovScore()));
        }

        // rank flow pool recall
        List<RecallResult.RecallData> flow_recall_rank = param.getRecallResult().getFlowPoolRecall().stream()
                .collect(Collectors.toList());
        Collections.sort(flow_recall_rank, Comparator.comparingDouble(o -> -o.getRovScore()));


        // TODO 重构
        //    去重原则：
        //        如果视频在ROV召回池topK，则保留ROV召回池，否则保留流量池
        Set<Long> rovTopKVideoIds = new HashSet<>();
        for (int i = 0; i < param.getTopK() && i < rov_recall_rank.size(); i++) {
            rovTopKVideoIds.add(rov_recall_rank.get(i).getVideoId());
        }

        Set<Long> flowPoolVideoIds = new HashSet<>();
        Iterator<RecallResult.RecallData> flowRecallRankIte = flow_recall_rank.iterator();
        while (flowRecallRankIte.hasNext()) {
            RecallResult.RecallData data = flowRecallRankIte.next();
            if (rovTopKVideoIds.contains(data.getVideoId())) {
                flowRecallRankIte.remove();
            } else {
                flowPoolVideoIds.add(data.getVideoId());
            }
        }

        Iterator<RecallResult.RecallData> rovRecallRankIte = rov_recall_rank.iterator();
        while (rovRecallRankIte.hasNext()) {
            RecallResult.RecallData data = rovRecallRankIte.next();
            if (flowPoolVideoIds.contains(data.getVideoId())) {
                rovRecallRankIte.remove();
            }
        }

        // 1 取topK
        if (CollectionUtils.isEmpty(rov_recall_rank)) {
            if (param.getSize() < flow_recall_rank.size()) {
                return new RankResult(flow_recall_rank.subList(0, param.getSize()));
            } else {
                return new RankResult(flow_recall_rank);
            }
        }

        // rov 取topK
        List<RecallResult.RecallData> datas = new ArrayList<>();
        for (int i = 0; i < param.getTopK() && i < rov_recall_rank.size(); i++) {
            datas.add(rov_recall_rank.get(i));
        }

        double flowPoolP = param.getFlowPoolP();
        if (param.getRecallResult().isQuickPool()) {
            String quick_flow_pool_P = redisTemplate.opsForValue().get("flow:pool:quick:distribute:rate:3");
            flowPoolP = NumberUtils.toDouble(quick_flow_pool_P, 0);
        }

        int flowPoolIndex = 0;
        int rovPoolIndex = param.getTopK() - 1;

        for (int i = 0; i < param.getSize() - param.getTopK(); i++) {
            double rand = RandomUtils.nextDouble(0, 1);
            if (rand < flowPoolP) {
                if (flowPoolIndex < flow_recall_rank.size()) {
                    datas.add(flow_recall_rank.get(flowPoolIndex++));
                } else {
                    break;
                }
            } else {
                if (rovPoolIndex < rov_recall_rank.size()) {
                    datas.add(rov_recall_rank.get(rovPoolIndex++));
                } else {
                    break;
                }
            }
        }
        //     while i < size - top_K:
        //        # 随机生成[0, 1)浮点数
        //        rand = random.random()
        //        # log_.info('rand: {}'.format(rand))
        //        if rand < flow_pool_P:
        //            if flow_recall_rank:
        //                rank_result.append(flow_recall_rank[0])
        //                flow_recall_rank.remove(flow_recall_rank[0])
        //            else:
        //                rank_result.extend(rov_recall_rank[:size - top_K - i])
        //                return rank_result[:size], flow_num
        //        else:
        //            if rov_recall_rank:
        //                rank_result.append(rov_recall_rank[0])
        //                rov_recall_rank.remove(rov_recall_rank[0])
        //            else:
        //                rank_result.extend(flow_recall_rank[:size - top_K - i])
        //                return rank_result[:size], flow_num
        //        i += 1

        if (param.getSize() < datas.size()) {
            return new RankResult(datas.subList(0, param.getSize()));
        }
        return new RankResult(datas);
    }
}
